package dk.nsp.xdscleanup.integrations.personinformation;



import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import dk.nsp.xdscleanup.service.ServiceErrorCallback;
import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.client.utils.URIBuilder;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.net.URISyntaxException;
import java.time.*;
import java.time.format.DateTimeFormatter;

import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class PersonInformationClient  {
    private static final String APPLICATION_JSON = "application/json";
    private static final String CONTENT_TYPE = "Content-Type";
    private static final Logger LOGGER = Logger.getLogger(PersonInformationClient.class);
    private final ObjectMapper objectMapper = new ObjectMapper();
    private final ServiceErrorCallback serviceErrorCallback;
    private final HttpClient httpClient;
    private final String url;

    public static final String personInformationServiceCallFailed = "Kaldet til PersonInformationService er fejlet.";
    public static final String personInformationServiceCallContenTypeWrong = "Kaldet til PersonInformation-servicen returnerede uventet contentType.";
    public static final String personInformationServiceCallError404 = "Fejl i retursvar fra PersonInformation. (404) forkert contentType";
    public static final String personInformationServiceCallParseError = "Kaldet til PersonInformation-servicen returnerede en uventet statuskode";
    public static final String personInformationServiceCallUnknownStatus = "Kaldet til PersonInformation-servicen returnerede en uventet statuskode: %d";
    public static final String personInformationServiceCallRequestError = "Opret af kald til PersonInformation-servicen gik galt.";
    public PersonInformationClient(HttpClient httpClient, String url, ServiceErrorCallback serviceErrorCallback) {
        this.httpClient = httpClient;
        this.url = url;
        this.serviceErrorCallback = serviceErrorCallback;
    }

    public PersonInformationClient(HttpClient httpClient, String url, Duration errorCountDuration, ZoneId zoneId) {
        this(httpClient, url, new ServiceErrorCallback(errorCountDuration, zoneId));
    }

    public List<String>  personDeceased(List<String> cprs, Instant deceasedSince) throws PersonInformationClientException {
        HttpUriRequest cprRequest = buildGetDeceasedRequest(cprs, deceasedSince);
        List<String> response = invokeService(cprRequest);
        return response;
    }

    public int getErrorCount() {
        return serviceErrorCallback.getErrorCount();
    }

    private HttpUriRequest buildGetDeceasedRequest(List<String> cprs, Instant deceasedSince) throws PersonInformationClientException {
        String utcDateString = DateTimeFormatter.ISO_INSTANT.format(deceasedSince);
        String urlWithContext = String.format("%s/person/deceased?", url);
        try {
            URIBuilder builder = new URIBuilder(urlWithContext);
            for (String cpr : cprs) {
                builder.addParameter("cpr", cpr);
            }
            builder.addParameter("dato", utcDateString);
            return new HttpGet(builder.build());
        } catch (URISyntaxException e) {
            String error = personInformationServiceCallRequestError;
            LOGGER.error(error);
            throw new PersonInformationClientException(error, e);
        }
    }

    private List<String> invokeService(HttpUriRequest request) throws PersonInformationClientException {
        try {
            HttpResponse response = httpClient.execute(request);
            List<String> result = handleResponse(response);
            serviceErrorCallback.registerServiceSuccess();
            return result;
        } catch (IOException e) {
            String errorMessage = personInformationServiceCallFailed;
            LOGGER.error(errorMessage, e);
            serviceErrorCallback.registerServiceError();
            throw new PersonInformationClientException(errorMessage, e);
        } catch (PersonInformationClientException e) {
            serviceErrorCallback.registerServiceError();
            throw e;
        }
    }

    private List<String> handleResponse(HttpResponse response) throws PersonInformationClientException {
        int statusCode = response.getStatusLine().getStatusCode();
        switch (statusCode) {
            case HttpStatus.SC_OK: {
                boolean contentTypeOk = contentTypeOk(response);
                List<String> responseData = parseResponseAsJsonList(response);
                if (!contentTypeOk) {
                    String error = personInformationServiceCallContenTypeWrong;
                    LOGGER.error(error);
                    throw new PersonInformationClientException(error);
                }
                return responseData;
            }
            case HttpStatus.SC_NOT_FOUND: {
                boolean contentTypeOk = contentTypeOk(response);
                if (!contentTypeOk) {
                    String error = personInformationServiceCallError404;
                    throw new PersonInformationClientException(error);
                }
                return parseResponseAsJsonList(response);
            }
            default:
                throw new PersonInformationClientException(String.format(personInformationServiceCallUnknownStatus, statusCode));
        }
    }

    private Map<String, Object> parseResponseAsJsonMap(HttpResponse response) throws PersonInformationClientException {
        try {
            return objectMapper.readValue(response.getEntity().getContent(), Map.class);
        } catch (IOException e) {
            LOGGER.error("Kunne ikke parse response body til json", e);
            throw new PersonInformationClientException(personInformationServiceCallParseError, e);
        }
    }
    private List<String> parseResponseAsJsonList(HttpResponse response) throws PersonInformationClientException {
        try {
            List<Map<String, String>> list = objectMapper.readValue(response.getEntity().getContent(), new TypeReference<List<Map<String, String>>>(){});
            List<String> result = list.stream()
                    .map(map -> map.get("cpr"))
                    .collect(Collectors.toList());
            return result;
        } catch (IOException e) {
            LOGGER.error("Kunne ikke parse response body til json", e);
            throw new PersonInformationClientException(personInformationServiceCallParseError, e);
        }
    }

    private boolean contentTypeOk(HttpResponse response) {
        Header headers = response.getFirstHeader(CONTENT_TYPE);
        return headers != null && headers.getValue().equalsIgnoreCase(APPLICATION_JSON);
    }

}
