package dk.nsp.xdscleanup.integrations.personinformation;

import dk.nsp.xdscleanup.service.ServiceErrorCallback;


import org.apache.http.HttpResponse;
import org.apache.http.StatusLine;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.entity.BasicHttpEntity;
import org.apache.http.message.BasicHttpResponse;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.mockito.Mockito;
import static org.junit.Assert.*;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.util.Arrays;
import java.util.List;

import static org.mockito.Mockito.*;

public class PersonInformationClientTest {

    private HttpClient httpClient;
    private final String url = "test-url";
    private ServiceErrorCallback serviceErrorCallback;
    private PersonInformationClient subject;

    @Rule
    public ExpectedException exceptionRule = ExpectedException.none();

    private final List<String> cprs = Arrays.asList(new String[]{"100001", "100002", "100003"});

    @Before
    public void setUp() {
        httpClient = mock(HttpClient.class);
        serviceErrorCallback = mock(ServiceErrorCallback.class);
        subject = new PersonInformationClient(httpClient, url, serviceErrorCallback);
    }

    @Test
    public void personDeceased_WhenServiceReturnsValidResponse_ReturnsCprsAndRegistersServiceSuccess() throws PersonInformationClientException, IOException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(200, "application/json",
                "[{\"cpr\":\"100001\"},{\"cpr\":\"100002\"},{\"cpr\":\"100003\"}]");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        // Act
        List<String> result = subject.personDeceased(cprs, Instant.now());

        // Assert
        assertEquals(3, result.size());
        assertEquals(cprs, result);
        verify(serviceErrorCallback).registerServiceSuccess();
    }

    @Test
    public void personDeceased_WhenClientThrowsIOException_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException, PersonInformationClientException {
        // Arrange
        doThrow(IOException.class).when(httpClient).execute(any(HttpUriRequest.class));
        exceptionRule.expect(PersonInformationClientException.class);
        exceptionRule.expectMessage(PersonInformationClient.personInformationServiceCallFailed);

        // Act
        subject.personDeceased(cprs, Instant.now());

        // Assert
        // exception
    }

    @Test
    public void personDeceased_WhenServiceReturnsResponse200WithContentTypeText_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException, PersonInformationClientException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(200, "text/plain",
                "[{\"cpr\":\"100001\"},{\"cpr\":\"100002\"},{\"cpr\":\"100003\"}]");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));
        exceptionRule.expect(PersonInformationClientException.class);
        exceptionRule.expectMessage(PersonInformationClient.personInformationServiceCallContenTypeWrong);

        // Act
        subject.personDeceased(cprs, Instant.now());

        // Assert
        // exception
    }

    @Test
    public void personDeceased_WhenServiceReturnsResponse200WithInvalidJson_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException, PersonInformationClientException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(200, "application/json", "invalid body");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));
        exceptionRule.expect(PersonInformationClientException.class);
        exceptionRule.expectMessage(PersonInformationClient.personInformationServiceCallParseError);

        // Act
        subject.personDeceased(cprs, Instant.now());

        // Assert
        // exception
    }


    @Test
    public void personDeceased_WhenServiceReturnsResponse404WithContentTypeText_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException, PersonInformationClientException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(404, "text/plain", "[{\"cpr\":\"100001\"},{\"cpr\":\"100002\"},{\"cpr\":\"100003\"}]");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));
        exceptionRule.expect(PersonInformationClientException.class);
        exceptionRule.expectMessage(PersonInformationClient.personInformationServiceCallError404);

        // Act
        subject.personDeceased(cprs, Instant.now());

        // Assert
        // exception

    }

    @Test
    public void personDeceased_WhenServiceReturnsResponse404WithInvalidJson_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException, PersonInformationClientException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(404, "application/json", "invalid body");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));
        exceptionRule.expect(PersonInformationClientException.class);
        exceptionRule.expectMessage(PersonInformationClient.personInformationServiceCallParseError);

        // Act
        subject.personDeceased(cprs, Instant.now());

        // Assert
        // exception
    }

    @Test
    public void personDeceased_WhenServiceReturnsResponse404WithValidJson_ReturnsFalseAndRegistersServiceSuccess() throws PersonInformationClientException, IOException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(404, "application/json", "[{\"cpr\":\"100001\"},{\"cpr\":\"100002\"},{\"cpr\":\"100003\"}]");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));

        // Act
        List<String> result = subject.personDeceased(cprs, Instant.now());

        // Assert
        assertEquals(3, result.size());
        assertEquals(cprs, result);
        verify(serviceErrorCallback).registerServiceSuccess();
    }

    @Test
    public void personDeceased_WhenServiceReturnsResponse400_ThrowsPersonInformationClientExceptionAndRegistersError() throws IOException, PersonInformationClientException {
        // Arrange
        HttpResponse httpResponse = createHttpResponse(400, "application/json", "{\"test\": \"test\"}");
        doAnswer(arg -> httpResponse).when(httpClient).execute(any(HttpUriRequest.class));
        exceptionRule.expect(PersonInformationClientException.class);
        exceptionRule.expectMessage(String.format(PersonInformationClient.personInformationServiceCallUnknownStatus, 400));

        // Act
        subject.personDeceased(cprs, Instant.now());

        // Assert
        // exception
    }

    private HttpResponse createHttpResponse(int statusCode, String contentType, String body) {
        StatusLine statusLine = Mockito.mock(StatusLine.class);
        Mockito.when(statusLine.getStatusCode()).thenReturn(statusCode);
        BasicHttpResponse httpResponse = new BasicHttpResponse(statusLine);
        BasicHttpEntity httpEntity = new BasicHttpEntity();
        httpEntity.setContent(inputStreamFromString(body));
        httpResponse.setEntity(httpEntity);
        httpResponse.setHeader("Content-Type", contentType);
        return httpResponse;
    }

    private static InputStream inputStreamFromString(String input) {
        return new ByteArrayInputStream(input.getBytes(StandardCharsets.UTF_8));
    }

}