package dk.nsp.xdscleanup.setup;

import dk.nsp.backgroundjob.service.StackedOperationsService;
import dk.nsp.backgroundjob.util.DateTimeProvider;
import dk.nsp.xdscleanup.factory.CleanupOperationsFactory;
import dk.nsp.xdscleanup.integrations.personinformation.PersonInformationClient;
import dk.nsp.xdscleanup.model.DocumentTypeCleanupSpecification;
import dk.nsp.xdscleanup.nxrg.dao.NxrgDao;
import dk.nsp.xdscleanup.nxrg.service.RegistryDeletionService;
import dk.nsp.xdscleanup.openxds.dao.OpenXdsDao;
import dk.nsp.xdscleanup.openxds.service.RepositoryDeletionService;
import dk.nsp.xdscleanup.stack.registry.deceased.PersonInformationService;
import dk.nsp.xdscleanup.stack.registry.deceased.DeceasedRegistryCleanupSupplier;
import dk.nsp.xdscleanup.stack.registry.timebased.TypeSpecificationBasedRegistryCleanupSupplier;
import dk.nsp.xdscleanup.stack.repository.RepositoryCleanupSupplier;
import dk.nsp.xdscleanup.util.PoolingHttpClientFactory;
import org.apache.http.client.HttpClient;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Import;
import org.springframework.context.annotation.PropertySource;
import org.springframework.context.support.PropertySourcesPlaceholderConfigurer;

import java.time.Duration;
import java.time.Period;
import java.time.ZoneId;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

@Configuration
@PropertySource("classpath:xdscleanup.properties")
@Import({DBSetup.class, VersionSetup.class})
public class XdsCleanupSetup {
    @Value("${deletion.registry.batchsize}")
    private int batchSizeRegistry;

    @Value("${deletion.repository.batchsize}")
    private int batchSizeRepository;

    @Value("${deletion.registry.timebased.limit}")
    private int registryTimebasedLimit;

    @Value("${deletion.repository.limit}")
    private int repositoryLimit;

    @Value("${deletion.save.deceased.docs}")
    private String deletionSaveDeceasedDocs;

    @Value("${desired.execution.duration}")
    private String desiredExecutionDuration;

    @Value("${dk.nsp.xdscleanup.httpclient.pooling.totalconnections}")
    private int maxTotalConnections;

    @Value("${dk.nsp.xdscleanup.httpclient.pooling.maxconnections.pr.route}")
    private int defaultMaxConnectionsPerRoute;

    @Value("${personinformation.url}")
    private String personInformationUrl;

    @Value("${personinformation.errorcount.duration:PT10M}")
    private String personInformationErrorCountDuration;

    @Value("${personinformation.error.tolerance:0}")
    private int personInformationErrorTolerance;

    @Bean
    public static PropertySourcesPlaceholderConfigurer propertySourcesPlaceholderConfigurer() {
        return new PropertySourcesPlaceholderConfigurer();
    }

    @Bean
    public RepositoryDeletionService repositoryDeletionService(OpenXdsDao openXdsDao) {
        return new RepositoryDeletionService(openXdsDao);
    }

    @Bean
    public RegistryDeletionService registryDeletionService(NxrgDao nxrgDao) {
        return new RegistryDeletionService(nxrgDao);
    }

    @Bean
    public StackedOperationsService registryTimeBasedCleanupService(List<DocumentTypeCleanupSpecification> documentTypeCleanupSpecifications,
                                                                          CleanupOperationsFactory cleanupOperationsFactory, Duration desiredExecutionDuration,
                                                                          DateTimeProvider dateTimeProvider) {

        TypeSpecificationBasedRegistryCleanupSupplier defaultStackOperation = new TypeSpecificationBasedRegistryCleanupSupplier(documentTypeCleanupSpecifications, cleanupOperationsFactory);
        return new StackedOperationsService(defaultStackOperation, desiredExecutionDuration, dateTimeProvider).withServiceName("Registry Time Based Cleanup");
    }

    @Bean
    public StackedOperationsService registryDeceasedCleanupService(CleanupOperationsFactory cleanupOperationsFactory, Duration desiredExecutionDuration,
                                                                         DateTimeProvider dateTimeProvider) {
        DeceasedRegistryCleanupSupplier defaultStackOperation = new DeceasedRegistryCleanupSupplier(cleanupOperationsFactory);
        return new StackedOperationsService(defaultStackOperation, desiredExecutionDuration, dateTimeProvider).withServiceName("Registry Deceased Cleanup");
    }

    @Bean
    public StackedOperationsService repositoryCleanupService(RegistryDeletionService registryDeletionService, CleanupOperationsFactory cleanupOperationsFactory,
                                                                   Duration desiredExecutionDuration, DateTimeProvider dateTimeProvider) {
        RepositoryCleanupSupplier defaultStackOperation = new RepositoryCleanupSupplier(registryDeletionService, cleanupOperationsFactory, batchSizeRepository, repositoryLimit);
        return new StackedOperationsService(defaultStackOperation, desiredExecutionDuration, dateTimeProvider).withServiceName("Repository Cleanup");
    }

    @Bean
    public CleanupOperationsFactory cleanupOperationsFactory(RegistryDeletionService registryDeletionService, RepositoryDeletionService repositoryDeletionService,
                                                             DateTimeProvider dateTimeProvider, Period deceasedPeriod, PersonInformationService personInformationService) {
        return new CleanupOperationsFactory(registryDeletionService, repositoryDeletionService, dateTimeProvider, batchSizeRegistry, deceasedPeriod, registryTimebasedLimit, personInformationService);
    }

    @Bean
    public DateTimeProvider dateTimeProvider() {
        return new DateTimeProvider();
    }

    @Bean
    public List<DocumentTypeCleanupSpecification> getDocumentTypeCleanupSpecifications(@Value("${deletion.document.cleanup.specifications}") String[] cleanupSpecifications) {
        return Arrays.stream(cleanupSpecifications)
                .map(DocumentTypeCleanupSpecification::parse)
                .collect(Collectors.toList());
    }

    @Bean
    public Period deceasedPeriod() {
        return Period.parse("P" + deletionSaveDeceasedDocs);
    }

    @Bean
    public Duration desiredExecutionDuration() {
        return Duration.parse(desiredExecutionDuration);
    }

    @Bean
    public PoolingHttpClientFactory poolingHttpClientFactory() {
        return new PoolingHttpClientFactory(maxTotalConnections, defaultMaxConnectionsPerRoute);
    }
    @Bean
    public PersonInformationClient personInformationClient(PoolingHttpClientFactory poolingHttpClientFactory, ZoneId zoneId) {
        HttpClient httpClient = poolingHttpClientFactory.getClient();
        return new PersonInformationClient(httpClient, personInformationUrl, Duration.parse(personInformationErrorCountDuration), zoneId);
    }

    @Bean
    public PersonInformationService personInformationService(PersonInformationClient personInformationClient) {
        return new PersonInformationService(personInformationClient, personInformationErrorTolerance);
    }

    @Bean
    public ZoneId zoneId() {
        return ZoneId.of("Europe/Copenhagen");
    }

}
