package dk.nsp.xdscleanup.stack.registry.deceased;

import dk.nsp.backgroundjob.stack.StackOperation;
import dk.nsp.backgroundjob.util.DateTimeProvider;
import dk.nsp.xdscleanup.common.model.DocumentEntry;
import dk.nsp.xdscleanup.factory.CleanupOperationsFactory;
import dk.nsp.xdscleanup.nxrg.service.RegistryDeletionService;
import dk.nsp.xdscleanup.stack.registry.RegistryBatchCleanupOperation;
import dk.nsp.xdscleanup.common.util.CleanupException;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;

import java.time.Period;
import java.time.ZonedDateTime;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import static dk.nsp.xdscleanup.TestUtils.assertThrows;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public class DeceasedRegistryPatientIdBatchCleanupSupplierTest {

    private RegistryDeletionService registryDeletionService;
    private DateTimeProvider dateTimeProvider;
    private CleanupOperationsFactory cleanupOperationsFactory;
    private PersonInformationService personInformationService;

    @Before
    public void setUp() throws Exception {
        registryDeletionService = Mockito.mock(RegistryDeletionService.class);
        dateTimeProvider = Mockito.mock(DateTimeProvider.class);
        cleanupOperationsFactory = Mockito.mock(CleanupOperationsFactory.class);
        personInformationService = Mockito.mock(PersonInformationService.class);

    }

    @Test
    public void run_WithNoDocumentsForCitizens_ReturnsEmptyList() {
        List<String> citizenCprs = Collections.singletonList("1234567890");
        DeceasedRegistryPatientIdBatchCleanupSupplier deceasedRegistryPatientIdBatchCleanupSupplier = new DeceasedRegistryPatientIdBatchCleanupSupplier(
                citizenCprs, registryDeletionService, dateTimeProvider, cleanupOperationsFactory, 1, Period.ofMonths(1), personInformationService);
        Mockito.when(dateTimeProvider.now()).thenReturn(ZonedDateTime.now());
        Mockito.when(registryDeletionService.getDocumentEntriesForCitizens(Mockito.anyList())).thenReturn(Collections.emptyList());

        List<StackOperation> stackOperations = deceasedRegistryPatientIdBatchCleanupSupplier.run();

        assertEquals(0, stackOperations.size());
    }

    @Test
    public void run_With3DocumentsForCitizensBatchSize1_Returns3StackOperationWithDocumentEntries() {
        List<String> citizenCprs = Arrays.asList("0000000000", "1111111111", "2222222222");
        Period deceasedPeriod = Period.ofMonths(1);
        DeceasedRegistryPatientIdBatchCleanupSupplier deceasedRegistryPatientIdBatchCleanupSupplier = new DeceasedRegistryPatientIdBatchCleanupSupplier(
                citizenCprs, registryDeletionService, dateTimeProvider, cleanupOperationsFactory, 1, deceasedPeriod, personInformationService);
        ZonedDateTime now = ZonedDateTime.now();
        Mockito.when(dateTimeProvider.now()).thenReturn(now);
        List<String> deceasedCprs = Arrays.asList("1111111111", "2222222222");
        Mockito.when(personInformationService.filterToDeceased(citizenCprs, now.minus(deceasedPeriod).toInstant()))
                .thenReturn(deceasedCprs);
        DocumentEntry entry1 = Mockito.mock(DocumentEntry.class);
        DocumentEntry entry2 = Mockito.mock(DocumentEntry.class);
        DocumentEntry entry3 = Mockito.mock(DocumentEntry.class);
        List<DocumentEntry> documentEntries = Arrays.asList(entry1, entry2, entry3);
        Mockito.when(registryDeletionService.getDocumentEntriesForCitizens(deceasedCprs)).thenReturn(documentEntries);
        RegistryBatchCleanupOperation batchOperation1 = Mockito.mock(RegistryBatchCleanupOperation.class);
        Mockito.when(batchOperation1.withOperationName(Mockito.anyString())).thenReturn(batchOperation1);
        Mockito.when(cleanupOperationsFactory.createRegistryBatchCleanupOperation(Arrays.asList(entry1))).thenReturn(batchOperation1);
        RegistryBatchCleanupOperation batchOperation2 = Mockito.mock(RegistryBatchCleanupOperation.class);
        Mockito.when(batchOperation2.withOperationName(Mockito.anyString())).thenReturn(batchOperation2);
        Mockito.when(cleanupOperationsFactory.createRegistryBatchCleanupOperation(Arrays.asList(entry2))).thenReturn(batchOperation2);
        RegistryBatchCleanupOperation batchOperation3 = Mockito.mock(RegistryBatchCleanupOperation.class);
        Mockito.when(batchOperation3.withOperationName(Mockito.anyString())).thenReturn(batchOperation3);
        Mockito.when(cleanupOperationsFactory.createRegistryBatchCleanupOperation(Arrays.asList(entry3))).thenReturn(batchOperation3);

        List<StackOperation> stackOperations = deceasedRegistryPatientIdBatchCleanupSupplier.run();

        assertEquals(3, stackOperations.size());
        assertTrue(stackOperations.contains(batchOperation1));
        assertTrue(stackOperations.contains(batchOperation2));
        assertTrue(stackOperations.contains(batchOperation3));
    }

    @Test
    public void run_WhenRetrievingDeceasedCitizensFails_ThrowsCleanupException() {
        Mockito.when(dateTimeProvider.now()).thenReturn(ZonedDateTime.now());
        Mockito.when(personInformationService.filterToDeceased(Mockito.anyList(), Mockito.any())).thenThrow(RuntimeException.class);
        DeceasedRegistryPatientIdBatchCleanupSupplier deceasedRegistryPatientIdBatchCleanupSupplier = new DeceasedRegistryPatientIdBatchCleanupSupplier(
                Collections.emptyList(), registryDeletionService, dateTimeProvider, cleanupOperationsFactory, 1, Period.ofMonths(1), personInformationService);

        CleanupException e = assertThrows(CleanupException.class, () -> deceasedRegistryPatientIdBatchCleanupSupplier.run());

        assertTrue(e.getMessage(), e.getMessage().contains("Failed to retrieve deceased citizens with latestStatusTime"));
    }

    @Test
    public void run_WhenRetrievingDocumentEntriesFails_ThrowsCleanupException() {
        Mockito.when(dateTimeProvider.now()).thenReturn(ZonedDateTime.now());
        Mockito.when(registryDeletionService.getDocumentEntriesForCitizens(Mockito.anyList())).thenThrow(RuntimeException.class);
        DeceasedRegistryPatientIdBatchCleanupSupplier deceasedRegistryPatientIdBatchCleanupSupplier = new DeceasedRegistryPatientIdBatchCleanupSupplier(
                Collections.emptyList(), registryDeletionService, dateTimeProvider, cleanupOperationsFactory, 1, Period.ofMonths(1), personInformationService);

        CleanupException e = assertThrows(CleanupException.class, () -> deceasedRegistryPatientIdBatchCleanupSupplier.run());

        assertEquals("Failed to retrieve document entries for citizens", e.getMessage());
    }
}